"""Library of GangaRobot.Framework.Action.IAction implementations.

See Base for abstract submit / finish / extract / report implementations.
See Core for generic submit / finish / extract / report implementations.
See Ext for specialised implementations.
See Example for example implementations.

"""