"""Generic submit / finish / extract / report action implementations.

See CoreSubmitter, CoreFinisher, CoreExtractor and CoreReporter for core
implementations providing generic submit, finish, extract and report actions by
extending the base implementations in GangaRobot.Lib.Base.

Specific action implementations can reuse or extend these core implementations.

See GangaRobot.Lib.Example for implementations which extend these reuse these
core implementations.

"""